@@ -4,7 +4,7 @@ import logging
 import traceback
 import threading
 import multiprocessing
-from time import time, sleep
+from time import sleep
 
 from setproctitle import setproctitle
 
@@ -38,18 +38,20 @@ class Worker(multiprocessing.Process):
         it will be invoked when worker.start() is called.
 
         """
+        logger.debug('Process id: %s', os.getpid())
+        logger.debug('Process group id: %s', os.getpgrp())
         setproctitle("kuyruk: worker on %s" % self.queue_name)
-        self.start_wathcing_master()
         self.register_signals()
-        self.started = time()
         self.queue.declare()
         self.channel.basic_qos(prefetch_count=1)
         # self.channel.tx_select()
 
+        self.start_wathcing_master()
+        if self.config.MAX_RUN_TIME > 0:
+            self.start_shutdown_timer()
+
         logger.info('Starting consume')
         for tag, task_description in self.queue:
-            if self.should_quit():
-                break
             self.on_task(tag, task_description)
 
         logger.debug("End run worker")
@@ -68,7 +70,7 @@ class Worker(multiprocessing.Process):
     def process_task(self, tag, task_description):
         from kuyruk import Kuyruk
         try:
-            self.import_and_call_task(task_description)
+            self.process_data_events_and_run_task(task_description)
         # sleep() calls below prevent cpu burning
         except Kuyruk.Reject:
             logger.warning('Task is rejected')
@@ -83,6 +85,21 @@ class Worker(multiprocessing.Process):
             logger.info('Task is successful')
             self.queue.ack(tag)
 
+    def process_data_events_and_run_task(self, task_description):
+        """Start a thread in background that listens data events from
+        connection for keeping it open and call task function in main thread.
+
+        """
+        try:
+            self.dep = DataEventProcessor(self.channel.connection)
+            self.dep.start()
+            self.import_and_call_task(task_description)
+        except Exception:
+            raise
+        finally:
+            self.dep.stop()
+            self.dep.join()
+
     def handle_exception(self, tag, task_description):
         retry_count = task_description.get('retry', 0)
         if retry_count:
@@ -132,16 +149,6 @@ class Worker(multiprocessing.Process):
         result = task.run(args, kwargs)
         logger.debug('Result: %r', result)
 
-    def should_quit(self):
-        def checks():
-            CHECKS = [
-                self.is_run_time_exceeded,
-                self.is_master_dead,
-            ]
-            for check in CHECKS:
-                yield check()
-        return any(checks())
-
     def is_master_dead(self):
         try:
             os.kill(os.getppid(), 0)
@@ -168,12 +175,21 @@ class Worker(multiprocessing.Process):
         t.daemon = True
         t.start()
 
-    def is_run_time_exceeded(self):
-        if self.config.MAX_RUN_TIME is not None:
-            passed_seconds = time() - self.started
-            if passed_seconds >= self.config.MAX_RUN_TIME:
-                logger.warning('Kuyruk run for %s seconds', passed_seconds)
-                return True
+    def start_shutdown_timer(self):
+        """
+        Start a Thread that counts down from MAX_RUN_TIME. When it reaches
+        zero it sends a signal to itself for graceful shutdown.
+
+        """
+        def watch():
+            sleep(self.config.MAX_RUN_TIME)
+            logger.critical('Run time reached zero, cancelling consume.')
+            # We do not call the handler directly here because
+            # pika is not thread safe.
+            os.kill(os.getpid(), signal.SIGTERM)
+        t = threading.Thread(target=watch)
+        t.daemon = True
+        t.start()
 
     def is_load_high(self):
         return os.getloadavg()[0] > self.config.MAX_LOAD
@@ -188,3 +204,23 @@ class Worker(multiprocessing.Process):
         logger.warning("Catched SIGTERM")
         logger.warning("Stopping %s...", self)
         self.queue.cancel()
+
+
+class DataEventProcessor(threading.Thread):
+
+    def __init__(self, connection):
+        super(DataEventProcessor, self).__init__()
+        self.daemon = True
+        self.connection = connection
+        self._stop = threading.Event()
+
+    def run(self):
+        while not self._stop.is_set():
+            try:
+                self.connection.process_data_events()
+            except Exception:
+                os._exit(1)
+            sleep(0.1)
+
+    def stop(self):
+        self._stop.set()
